@@ -1,13 +1,14 @@
-
+#include <stdexcept>
+#include "bigintegerR.h"
 #include <R.h>
 #include <Rinternals.h>
 // and one thing from Rdefines.h :
 #define NEW_LIST(n) allocVector(VECSXP,n)
 
 #include "apply.h"
-#include "bigintegerR.h"
+
 #include "bigrationalR.h"
-#include <stdexcept>
+
 
 // X a matrix, or a bigz integer
 // line: true = we return a list of all lines
@@ -66,7 +67,7 @@ SEXP gmpMatToListZ(SEXP X, SEXP line)
 
       return(ans);
     } catch(std::invalid_argument & e){
-    error(e.what());
+   error("%s",e.what());
   }
   
 }
@@ -126,7 +127,7 @@ SEXP gmpMatToListQ(SEXP X, SEXP line)
       return(ans);
 
     } catch(std::invalid_argument & e){
-    error(e.what());
+    error("%s",e.what());
   }
   
 }
